import typing


def cartesian_product(grid: typing.Dict):
    """
    Helper function to compute the cartesian product
    of an experiment grid.
    For instance the grid
        {
            a: ['a1', 'a2', 'a3'],
            b: ['b1', 'b2']
        }
    will produce
        [
            {a: 'a1', b:'b1'},
            {a: 'a2', b:'b2'},
            {a: 'a3', b:'b1'},
            {a: 'a1', b:'b2'},
            {a: 'a2', b:'b1'},
            {a: 'a3', b:'b2'}
        ]
    """
    my_keys = list(grid.keys())

    configs = [
        {my_keys[0]: val}
        for val in grid[my_keys[0]]
    ]

    print(configs)

    for key in my_keys[1:]:
        my_values = grid[key]
        new_configs = list()
        for config in configs:
            for value in my_values:
                new_configs.append({key: value, **config})
        configs = new_configs

    return configs


def namedtuple_to_dict(input: typing.NamedTuple) -> typing.Dict:
    root = input._asdict()
    my_stash = [root]
    while len(my_stash) > 0:
        subroot = my_stash.pop()
        for key in subroot:
            if hasattr(subroot[key], "_asdict"):
                subroot[key] = subroot[key]._asdict()
                my_stash.append(subroot[key])
    return root



import unittest
from functools import reduce
import typing
import json

from r3d3.utils import cartesian_product, namedtuple_to_dict


class FakeSubConfig(typing.NamedTuple):
    foo: str = "bar"


class FakeConfig(typing.NamedTuple):
    alpha: float = 0
    sub_conf: FakeSubConfig = FakeSubConfig()


class TestUtils(unittest.TestCase):

    def test_cartesian_product(self):
        my_grid = {
            "a": ["a1", "a2", "a3"],
            "b": ["b1", "b2"],
            "c": ["c1", "c2"]
        }

        all_variants = cartesian_product(my_grid)

        expected_nb_variants = reduce(lambda x, y: x * y, [len(v) for v in my_grid.values()])
        self.assertEqual(len(all_variants), expected_nb_variants)

    def test_namedtuple_to_json(self):
        self.assertEqual(
            json.dumps(namedtuple_to_dict(FakeConfig())),
            "{\"alpha\": 0, \"sub_conf\": {\"foo\": \"bar\"}}"
        )



import unittest
import os

test_rootpath = os.path.dirname(os.path.abspath(__file__))

from r3d3.experiment_launcher import main


class TestExperimentLauncher(unittest.TestCase):

    def test_main(self):
        launcher = main(f"{test_rootpath}/samples/sample_experiment.py")
        self.assertEqual(launcher.db.get_nb_experiments(), 4)



import unittest
import typing
import os

from r3d3.experiment_db import ExperimentDB
import tempfile


class FakeConfig(typing.NamedTuple):
    alpha: float = 0


class TestExperimentDB(unittest.TestCase):

    def test_nb_experiments(self):

        db_path = tempfile.mkstemp()[1]

        db = ExperimentDB(db_path=db_path)
        db.init_experiment_table(drop=True)

        self.assertEqual(db.get_nb_experiments(), 0)

        db.add_experiment(
            experiment_id=0,
            run_id=1,
            config=FakeConfig()
        )
        self.assertEqual(db.get_nb_experiments(), 1)
        df = db.list_all_experiments()
        self.assertEqual(len(df), 1)

        os.remove(db_path)



import argparse
import sys
import os
from concurrent.futures import ThreadPoolExecutor
import subprocess
from datetime import datetime
import time
import typing

from r3d3.utils import cartesian_product
from r3d3 import R3D3Experiment, ExperimentDB

root_dir = "{}/..".format(os.path.dirname(os.path.abspath(__file__)))


class ExperimentLauncher(object):

    def __init__(self, db_path):
        self.db = ExperimentDB(db_path)

    def run(self, binary: str, configs: typing.List, max_nb_processes: int):
        self.db.init_experiment_table()

        def launcher_with_environment(env, debug):
            def launch_command_line(command):
                tab = command.split()
                print("Executing {}".format(command))
                if not debug:
                    print(tab)
                    try:
                        myPopen = subprocess.Popen(
                            tab,
                            env=env,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
                        for l in myPopen.stderr:
                            print(l)
                    except subprocess.CalledProcessError as e:
                        print(e.output)

            return launch_command_line

        # Creating env for the runs
        env = os.environ.copy()
        print("Using env {}".format(env))

        nb_tests = len(configs)
        print("%d experiments to launch..." % nb_tests)

        # Creating executors with max nb processes from the config
        executor = ThreadPoolExecutor(max_workers=max_nb_processes)

        # Running the tests
        now = datetime.now()
        experiment_id = int(time.mktime(now.timetuple()))

        for run_id, parameter_set in enumerate(configs):
            # The python binary is available in sys.executable
            args = ["{} {}".format(sys.executable, f"{binary}")]
            for a in parameter_set:
                args.append("-" + a + " " + str(parameter_set[a]))

            # Passing launcher information to the experiment
            args.append("--max_nb_processes {}".format(min([max_nb_processes, nb_tests])))
            args.append(f"--experiment_id {experiment_id}")
            args.append(f"--run_id {run_id}")

            self.db.add_experiment(
                experiment_id=experiment_id,
                run_id=run_id,
                config=parameter_set
            )

            command = " ".join(args)
            executor.submit(launcher_with_environment(env, debug=False), command)


def main(experiment_file: str):
    print(experiment_file)

    variables = dict()
    with open(experiment_file) as f:
        exec(f.read(), variables)

    my_experiment: R3D3Experiment = variables["experiment"]

    my_launcher = ExperimentLauncher(my_experiment.db_path)
    my_configs = cartesian_product(my_experiment.configs)
    my_launcher.run(
        binary=my_experiment.binary,
        configs=my_configs,
        max_nb_processes=my_experiment.max_nb_processes
    )

    return my_launcher


def main_cli():
    parser = argparse.ArgumentParser(description='Experiment Launcher')
    parser.add_argument('--experiment_file', type=str)
    args = parser.parse_args()

    main(
        experiment_file=args.experiment_file
    )




#!/usr/bin/env python3

import setuptools
import os


def _read_reqs(relpath):
    fullpath = os.path.join(os.path.dirname(__file__), relpath)
    with open(fullpath) as f:
        return [s.strip() for s in f.readlines()
                if (s.strip() and not s.startswith("#"))]


_REQUIREMENTS_TXT = _read_reqs("requirements.txt")
_INSTALL_REQUIRES = [l for l in _REQUIREMENTS_TXT if "://" not in l]

setuptools.setup(
    name='r3d3',
    version='0.5',
    install_requires=_INSTALL_REQUIRES,
    dependency_links=[],
    data_files=[('.', ['requirements.txt'])],
    entry_points={
        'console_scripts': [
            'r3d3-xp = r3d3.experiment_launcher:main_cli'
        ],
    },
    packages=setuptools.find_packages())



import datetime
import json
import logging
import os
import sqlite3
import typing
from contextlib import contextmanager

import pandas as pd

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("[ExperimentDB]")


class ExperimentDB(object):

    def __init__(self, db_path):
        self.db_path = db_path

    @contextmanager
    def db_cursor(self):
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        try:
            yield cursor
        except Exception as e:
            logger.error("Error: {}".format(e))
            conn.rollback()
        else:
            conn.commit()

    # Initialize experiments table
    def init_experiment_table(self, drop=False):
        with self.db_cursor() as cur:
            if drop:
                cur.execute("DROP TABLE IF EXISTS experiments")

            cur.execute('''CREATE TABLE IF NOT EXISTS experiments
                         (
                         experiment_id integer,
                         run_id integer,
                         date text,
                         config text,
                         metrics text,
                         owner text,
                         PRIMARY KEY (experiment_id, run_id, owner)
                         )''')

    def get_nb_experiments(self):
        nb_experiments = 0

        with self.db_cursor() as cur:
            cur.execute("SELECT count(1) FROM experiments")
            nb_experiments = cur.fetchone()[0]

        return nb_experiments

    def add_experiment(self, experiment_id: int,
                       run_id: int,
                       config: typing.Dict):
        self.init_experiment_table(drop=False)

        with self.db_cursor() as cur:
            date = str(datetime.datetime.now().isoformat())

            cur.execute(f"""INSERT INTO experiments VALUES (
                {experiment_id},
                {run_id},
                '{date}',
                '{json.dumps(config)}',
                '',
                '{os.environ.get("USER", "unknown")}'
            )
            """)

    def update_experiment(self, experiment_id, run_id, metrics):
        with self.db_cursor() as cur:
            cur.execute(f"""
            UPDATE experiments
            SET metrics = '{json.dumps(metrics)}'
            WHERE run_id = '{run_id}' AND experiment_id = '{experiment_id}'""")

    def list_all_experiments(self):
        with self.db_cursor() as cur:
            ret = list()
            for row in cur.execute("SELECT * FROM experiments"):
                ret.append(row)

        return pd.DataFrame(
            data=ret,
            columns=[
                "experiment_id",
                "run_id",
                "date",
                "config",
                "metrics",
                "owner"])

    @staticmethod
    def parse_json(s):
        try:
            return json.loads(s)
        except:
            return dict()

    @staticmethod
    def recursive_get(d: typing.Dict, path: str):
        path_spl = path.split(".")
        res = d
        for elem in path_spl:
            res = res.get(elem, dict())

        if isinstance(res, dict) and len(res) == 0:
            return None

        return res

    def show_experiment(self, metarun_id: int, params: typing.Dict, metrics: typing.Dict):
        with self.db_cursor() as cur:
            ret = list()
            for row in cur.execute(f"SELECT * FROM experiments WHERE metarun_id = '{metarun_id}'"):
                ret.append(row)

        df = pd.DataFrame(
            data=ret,
            columns=[
                "experiment_id",
                "run_id",
                "date",
                "config",
                "metrics",
                "owner"])
        df["run_id"] = df["run_id"].apply(int)

        for name in params:
            df[name] = df["config"].apply(lambda s:
                                          ExperimentDB.recursive_get(
                                              ExperimentDB.parse_json(s), params[name]))
        for name in metrics:
            df[name] = df["metrics"].apply(lambda s:
                                           ExperimentDB.recursive_get(
                                               ExperimentDB.parse_json(s), metrics[name]))

        df.drop(columns=["experiment_id", "date", "metrics", "config", "owner"], inplace=True)

        return df
